
#include <iostream>
#include <memory>
#include <type_traits>
#include "make_function_map.h"

namespace hoge{

    // 基底class
    class B{
    public:
        virtual ~B(){}
        
        int get_id() const
        { return id; }
    private:
        int const id;
    
    protected:
        explicit B(int id)
            : id(id)
        {}
    };
    
    // 具象クラス
    class D1 : public B
    {
    public:
        D1() : B(100) {}
    };
    
    class D2 : public B
    {
    public:
        D2() : B(200) {}
    };

    // テンプレート関数 collision
    template<typename Tp1, typename Tp2>
    bool collision(Tp1 a, Tp2 b);
    
    // テンプレート関数の特殊化
    template<>
    bool collision<D1 const&, D1 const&>(D1 const& a, D1 const& b)
    {
        return true;
    }

    template<>
    bool collision<D1 const&, D2 const&>(D1 const& a, D2 const& b)
    {
        return false;
    }

    template<>
    bool collision<D2 const&, D2 const&>(D2 const& a, D2 const& b)
    {
        return true;
    }

    template<>
    bool collision<D2 const&, D1 const&>(D2 const& a, D1 const& b)
    {
        return false;
    }
    
    // 関数マップをつくるのに必要
    struct _collision
    {
        template<typename Tp1, typename Tp2>
        static bool
        execute(Tp1 a, Tp2 b)
        {
            return collision<Tp1, Tp2>(a, b);
        }
    };
    
    // 関数マップで分岐
    bool collision(B const& a, B const& b)
    {
        using sykes::multi_method::id_and_type;
        
        // 関数マップを作成
        auto static const m =
        sykes::multi_method::make_function_map
        <
            bool(hoge::B const&, hoge::B const&),
            hoge::_collision,
            sykes::dynamic_caster,
            id_and_type<int, 100, hoge::D1 const&>,
            id_and_type<int, 200, hoge::D2 const&>
        >();
        
        std::tuple<int, int> key(a.get_id(), b.get_id());
        
        return m.at(key)(a, b);
    }
}//

int main()
{
    std::unique_ptr<hoge::B>
        b1(new hoge::D1()),
        b2(new hoge::D2());
    
    std::cout << collision(*b1, *b1) << std::endl;
    std::cout << collision(*b1, *b2) << std::endl;
    std::cout << collision(*b2, *b1) << std::endl;
    std::cout << collision(*b2, *b2) << std::endl;
}


